import random
from datetime import datetime
from functools import wraps
from pathlib import Path
from typing import Awaitable, Literal, Union

import aiohttp
import pyproj
import pytz

from . import t

with open(Path(__file__).parent.joinpath('ua.txt'), encoding='utf-8') as f:
    USER_AGENTS = tuple(a.strip() for a in f.readline())

ERR_MESSAGES = {
    'api-error': {
        'tc': 'API 錯誤',
        'en': 'API Error',
    },
    'empty': {
        'tc': '沒有預報',
        'en': 'No Data',
    },
    'eos': {
        'tc': '服務時間已過',
        'en': 'Not in Service',
    },
    'ss-effect': {
        'tc': '特別車務安排',
        'en': 'Special Service in Effect',
    }
}

EPSG_TRANSFORMER = pyproj.Transformer.from_crs('epsg:2326', 'epsg:4326')


def ensure_session(func: Awaitable):
    @wraps(func)
    async def wrapper(*args, **kwargs):
        if kwargs.get('session') is not None:
            assert isinstance(kwargs['session'], aiohttp.ClientSession)
            return await func(*args, **kwargs)
        async with aiohttp.ClientSession() as s:
            return await func(*args, **{**kwargs, 'session': s})
    return wrapper


def dt_to_8601(dt: datetime) -> str:
    '''Convert a `datetime` instance to ISO-8601 formatted string.'''
    return dt.isoformat(sep='T', timespec='seconds')


def timestamp():
    return datetime.now().replace(tzinfo=pytz.timezone('Etc/GMT-8'))


def error_eta(message: Union[Literal['api-error', 'empty', 'eos', 'ss-effect'], str],
              ts: datetime = None,
              language: t.Language = 'tc'):
    return {
        'timestamp': dt_to_8601(ts or timestamp()),
        'message': ERR_MESSAGES.get(message, {}).get(language, message),
        'etas': None
    }


def ua_header():
    return {'User-Agent': random.choice(USER_AGENTS)}


async def search_location(name: str, session: aiohttp.ClientSession) -> tuple[str, str]:
    async with session.get(
            f'https://geodata.gov.hk/gs/api/v1.0.0/locationSearch?q={name}') as request:
        first = (await request.json())[0]
        return EPSG_TRANSFORMER.transform(first['y'], first['x'])


async def is_up_to_date(path: Path, url: str, session: aiohttp.ClientSession) -> bool:
    async with session.get(url) as request:
        return path.stat().st_mtime >= datetime.strptime((await request.text()).split('\n')[1], '%Y-%m-%d').timestamp()


# async def gtfs_route_match(transport: t.Transport,
#                            routes: t.Route,
#                            session: aiohttp.ClientSession) -> t.Route:
#     def similarity(a: str, b: str) -> float:
#         return SequenceMatcher(None, a, b).quick_ratio()

#     def clean_name(s: str) -> str:
#         return s.replace('(循環線)', '')\
#             .translate(str.maketrans('', '', string.punctuation))\
#             .upper()

#     def direction_aware(gtfs_services: list[dict]) -> bool:
#         pairs = []
#         for service in gtfs_services:
#             if ((service['dest'], service['orig']) in pairs):
#                 return True
#             pairs.append((service['orig'], service['dest']))
#         return False

#     routes = deepcopy(routes)

#     _gtfs_data = (await _gtfs_parser.gtfs_routes(session=session))
#     routes_gtfs = _gtfs_data['kmb'] | _gtfs_data['lwb']

#     for no, bound in routes.items():
#         if no not in routes_gtfs:
#             continue
#         for direction, services in bound.items():
#             if len(services) == 1 and len(routes_gtfs[no]) == 1:
#                 routes[no][direction][0]['gtfs_id'] = routes_gtfs[no][0]['id']
#                 continue

#             matched_ids = set()
#             is_multiple = len(routes_gtfs[no]) > 1

#             # same orig & dest (272A)
#             # Different ID in opposite dir (960X)
#             # Via (28B, 30X)
#             # Special orig & dest matched (296M)
#             for idx, service in enumerate(services):
#                 for service_gtfs in routes_gtfs[no]:
#                     if service_gtfs['id'] in matched_ids:
#                         continue

#                     if idx == 0 and not direction_aware(routes_gtfs[no]):
#                         routes[no][direction][0]['gtfs_id'] = routes_gtfs[no][0]['id']
#                         matched_ids.add(routes_gtfs[no][0]['id'])
#                         break

#                     if (similarity(clean_name(service['orig']['zh']), service_gtfs['orig']) >= 0.6
#                             and similarity(clean_name(service['dest']['zh']), service_gtfs['dest']) >= 0.6):
#                         routes[no][direction][idx]['gtfs_id'] = service_gtfs['id']
#                         matched_ids.add(service_gtfs['id'])
#                         break

#                     if not idx >= 1:
#                         continue

#                     if (similarity(clean_name(service['description']['zh']), clean_name(service_gtfs['orig'])) > 0
#                             or similarity(clean_name(service['description']['zh']), clean_name(service_gtfs['dest'])) > 0):
#                         routes[no][direction][idx]['gtfs_id'] = service_gtfs['id']
#                         matched_ids.add(service_gtfs['id'])
#                         break

#                 if 'gtfs_id' in routes[no][direction][idx]:
#                     continue

#                 for service_gtfs in routes_gtfs[no]:
#                     if service_gtfs['id'] in matched_ids:
#                         continue
#                     service_gtfs['orig'] = clean_name(service_gtfs['orig'])
#                     service_gtfs['dest'] = clean_name(service_gtfs['dest'])

#                     if (similarity(service['dest']['zh'], service_gtfs['orig']) >= 0.6
#                             and similarity(service['orig']['zh'], service_gtfs['dest']) >= 0.6):
#                         routes[no][direction][idx]['gtfs_id'] = service_gtfs['id']
#                         matched_ids.add(service_gtfs['id'])
#     return routes



from typing import Literal, Optional, TypedDict


Transport = Literal['ctb', 'kmb', 'lrt', 'lrtfeeder', 'nlb', 'mtr']

Language = Literal['tc', 'en']

Direction = Literal['outbound', 'inbound']


class Eta(TypedDict):
    class Extras(TypedDict):
        destinaion: Optional[str]
        varient: Optional[str]
        platform: Optional[str]
        car_length: Optional[int]

    eta: str
    is_arriving: bool
    is_scheduled: bool
    extras: Extras
    remark: Optional[str]


class Etas(TypedDict):
    timestamp: str
    message: Optional[str]
    etas: Optional[Eta]


class Route(TypedDict):
    class Service(TypedDict):
        id: str
        gtfs_id: Optional[str]
        description: Optional[str]
        orig: dict[Language, str]
        dest: dict[Language, str]

    outbound: list[Service]
    inbound: list[Service]


class Stop(TypedDict):
    id: str
    seq: int
    name: dict[Language, str]



import asyncio
import csv
from datetime import datetime
from typing import Generator, Optional

import aiohttp
import pytz

from . import t
from ._utils import dt_to_8601, ensure_session, error_eta, search_location


@ensure_session
async def routes(*, session: aiohttp.ClientSession) -> dict[str, t.Route]:
    routes_ = {}
    async with session.get(
            'https://opendata.mtr.com.hk/data/mtr_lines_and_stations.csv') as request:
        for row in csv.reader((await request.text('utf-8')).splitlines()[1:]):
            # column definition:
            #   route, direction, stopCode, stopID, stopTCName, stopENName, seq
            if not any(row):  # skip empty lines
                continue

            direction, _, branch = row[1].partition('-')
            if branch:
                # route with branch lines
                direction, branch = branch, direction  # e.g. LMC-DT
            direction = 'outbound' if direction == 'DT' else 'inbound'
            routes_.setdefault(row[0], {'inbound': [], 'outbound': []})

            if (row[6] == '1.00'):
                # origin
                routes_[row[0]][direction].append({
                    'id': (route_id := '_'.join(filter(None, (row[0], direction, branch)))),
                    'description': None,
                    'orig': {'en': row[5], 'tc': row[4]},
                    'dest': {}
                })
            else:
                # destination
                if len(routes_[row[0]][direction]) == 1:
                    routes_[row[0]][direction][0]['dest'] = {
                        'en': row[5], 'tc': row[4]
                    }
                else:
                    for idx, branch in enumerate(routes_[row[0]][direction]):
                        if branch['id'] != route_id:
                            continue
                        routes_[row[0]][direction][idx]['dest'] = {
                            'en': row[5], 'tc': row[4]
                        }
                        break
    return routes_


@ensure_session
async def stops(route_id: str, *, session: aiohttp.ClientSession) -> Generator[t.Stop, None, None]:
    # column definition:
    #   route, direction, stopCode, stopID, stopTCName, stopENName, seq
    async with session.get(
            'https://opendata.mtr.com.hk/data/mtr_lines_and_stations.csv') as request:
        if len(route_id := route_id.split('_')) > 2:
            stops_ = [stop for stop in csv.reader((await request.text('utf-8')).splitlines()[1:])
                      if stop[0] == route_id[0]
                      and stop[1] == f'{route_id[2]}-{"DT" if route_id[1] == "outbound" else "UT"}']
        else:
            stops_ = [stop for stop in csv.reader((await request.text('utf-8')).splitlines()[1:])
                      if stop[0] == route_id[0]
                      and stop[1] == ('DT' if route_id[1] == 'outbound' else 'UT')]

    if len(stops_) == 0:
        raise KeyError('route not exists')

    locations = await asyncio.gather(
        *[search_location(f'\u6e2f\u9435{s[4]}\u7ad9', session) for s in stops_])
    return ({
        'id': s[2],
        'seq': int(s[6].removesuffix('.00')),
        'name': {'tc': s[4], 'en': s[5]},
        'location': locations[i]
    } for i, s in enumerate(stops_))


@ ensure_session
async def etas(route_id: str,
               stop_id: str,
               language: t.Language = 'tc',
               *,
               session: aiohttp.ClientSession) -> t.Etas:
    route, direction, _ = route_id.split('_')
    direction = 'DOWN' if direction == 'outbound' else 'UP'

    async with session.get('https://rt.data.gov.hk/v1/transport/mtr/getSchedule.php',
                           params={
                               'line': route,
                               'sta': stop_id,
                               'lang': language
                           }) as request:
        response = await request.json()

    if len(response) == 0:
        return error_eta('api-error')
    if response.get('status', 0) == 0:
        if 'suspended' in response['message']:
            return error_eta(response['message'])
        if response.get('url') is not None:
            return error_eta('ss-effect')
        return error_eta('api-error')

    etas_ = []
    timestamp = datetime.fromisoformat(response['curr_time'])\
        .astimezone(pytz.timezone('Asia/Hong_kong'))

    for entry in response['data'][f'{route}-{stop_id}'].get(direction, []):
        eta_dt = datetime.fromisoformat(entry['time'])\
            .astimezone(pytz.timezone('Asia/Hong_kong'))
        etas_.append({
            'eta': dt_to_8601(eta_dt),
            'is_arriving': (eta_dt - timestamp).total_seconds() < 90,
            'is_scheduled': False,
            'extras': {
                'destination': entry['dest'],
                'varient': _varient_text(entry.get('route'), language),
                'platform': entry['plat'],
                'car_length': None
            },
            'remark': None,
        })

    if len(etas_) == 0:
        return error_eta('empty')
    return {
        'timestamp': dt_to_8601(timestamp),
        'message': None,
        'etas': etas_
    }


def _varient_text(val: Optional[str], language: t.Language):
    if val == 'RAC':
        return '\u7d93\u99ac\u5834' if language == 'tc' else 'Via Racecourse'
    return val



import importlib
import sys
from typing import Coroutine, Iterable

import aiohttp

from . import t


def routes(co: t.Transport,
           *,
           session: aiohttp.ClientSession = None) -> Coroutine[None, None, dict[str, t.Route]]:
    return importlib.import_module(f'.{co}', sys.modules[__name__].__package__)\
        .__dict__\
        .get('routes')(session=session)


def stops(co: t.Transport,
          route_id: str,
          *,
          session: aiohttp.ClientSession = None) -> Coroutine[None, None, Iterable[t.Stop]]:
    return importlib.import_module(f'.{co}', sys.modules[__name__].__package__)\
        .__dict__\
        .get('stops')(route_id, session=session)


def etas(co: t.Transport,
         route_id: str,
         stop_id: str,
         language: t.Language = 'tc',
         *,
         session: aiohttp.ClientSession = None) -> Coroutine[None, None, t.Etas]:
    return importlib.import_module(f'.{co}', sys.modules[__name__].__package__)\
        .__dict__\
        .get('etas')(route_id, stop_id, language, session=session)



import asyncio
from datetime import datetime

import aiohttp

from . import t
from ._utils import dt_to_8601, ensure_session, error_eta


@ensure_session
async def routes(*, session: aiohttp.ClientSession) -> dict[str, t.Route]:
    async def ends(r: dict, s: aiohttp.ClientSession):
        # pylint: disable=line-too-long
        async with s.get(f'https://rt.data.gov.hk/v2/transport/citybus/route-stop/ctb/{r["route"]}/inbound') as request:
            return r['route'], {
                'outbound': [{
                    'id': f'{r["route"]}_outbound_1',
                    'description': None,
                    'orig': {
                        'tc': r['orig_tc'],
                        'en': r['orig_en']
                    },
                    'dest': {
                        'tc': r['dest_tc'],
                        'en': r['dest_en']
                    },
                }],
                'inbound': [] if len((await request.json())['data']) == 0 else [{
                    'id': f'{r["route"]}_inbound_1',
                    'description': None,
                    'orig': {
                        'tc': r['dest_tc'],
                        'en': r['dest_en']
                    },
                    'dest': {
                        'tc': r['orig_tc'],
                        'en': r['orig_en']
                    },

                }]
            }

    async with session.get('https://rt.data.gov.hk/v2/transport/citybus/route/ctb') as request:
        return {d[0]: d[1]
                for d in await asyncio.gather(*[ends(r, session)
                                                for r in (await request.json())['data']])
                }


@ensure_session
async def stops(route_id: str, *, session: aiohttp.ClientSession) -> list[dict[str,]]:
    async def fetch(stop: dict, session: aiohttp.ClientSession):
        async with session.get(
                f'https://rt.data.gov.hk/v2/transport/citybus/stop/{stop["stop"]}') as request:
            detail = (await request.json())['data']
            return {
                'id': stop['stop'],
                'seq': int(stop['seq']),
                'name': {
                    'tc': detail.get('name_tc'),
                    'en': detail.get('name_en'),
                },
                'location': (detail['lat'], detail['long'])
            }

    # pylint: disable=line-too-long
    async with session.get(
            f'https://rt.data.gov.hk/v2/transport/citybus/route-stop/ctb/{"/".join(route_id.split("_")[:2])}') as request:
        data = await asyncio.gather(
            *[fetch(stop, session) for stop in (await request.json())['data']])

    if len(data) == 0:
        raise KeyError('route not exists')
    return data


@ensure_session
async def etas(route_id: str,
               stop_id: str,
               language: t.Language = 'tc',
               *,
               session: aiohttp.ClientSession) -> t.Etas:
    route, direction, _ = route_id.split('_')

    async with session.get(
            f'https://rt.data.gov.hk/v2/transport/citybus/eta/ctb/{stop_id}/{route}') as request:
        response = await request.json()

    if len(response) == 0 or response.get('data') is None:
        return error_eta('api-error')
    if len(response['data']) == 0:
        return error_eta('empty')

    etas_ = []
    timestamp = datetime.fromisoformat(response['generated_timestamp'])

    for eta in response['data']:
        if eta['dir'].lower() != direction[0]:
            continue
        if eta['eta'] == '':
            # 九巴時段
            etas_.append({
                'eta': None,
                'is_arriving': False,
                'is_scheduled': True,
                'extras': {
                    'destinaion': eta[f'dest_{language}'],
                    'varient': None,
                    'platform': None,
                    'car_length': None
                },
                'remark': eta[f'rmk_{language}'],
            })
        else:
            eta_dt = datetime.fromisoformat(eta['eta'])
            etas_.append({
                'eta': dt_to_8601(eta_dt),
                'is_arriving': (eta_dt - timestamp).total_seconds() < 60,
                'is_scheduled': True,
                'extras': {
                    'destinaion': eta[f'dest_{language}'],
                    'varient': None,
                    'platform': None,
                    'car_length': None
                },
                'remark': eta[f'rmk_{language}'],
            })

    return {
        'timestamp': dt_to_8601(timestamp),
        'message': None,
        'etas': etas_
    }


# @ensure_session
# async def routes(*, session: aiohttp.ClientSession):
#     # Stop ID of the same stop from different route will have the same ID,
#     # caching the stop details to reduce the number of requests (around 600 - 700).
#     # Execution time is not guaranteed to be reduced.
#     stop_cache = {}
#     semaphore = asyncio.Semaphore(10)

#     async with session.get('https://rt.data.gov.hk/v2/transport/citybus/route/ctb') as response:
#         tasks = [_stop_list(s['route'], stop_cache, semaphore, session)
#                  for s in (await response.json())['data']]
#     return {route[0]: route[1] for route in await asyncio.gather(*tasks)}


# async def _route_ends(route: str,
#                       direction: Literal['inbound', 'outbound'],
#                       session: aiohttp.ClientSession) -> Optional[tuple[str, str]]:
#     # pylint: disable=line-too-long
#     async with session.get(
#             f'https://rt.data.gov.hk/v2/transport/citybus/route-stop/ctb/{route}/{direction}') as request:
#         data = (await request.json())['data']
#         return None if len(data) == 0 else (data[0]['stop'], data[-1]['stop'])


# async def _stop_list(
#         route: str, cache: dict, semaphore: asyncio.Semaphore, session: aiohttp.ClientSession):
#     async def _stop_name(stop_id: str, session: aiohttp.ClientSession) -> dict[str, str]:
#         async with session.get(
#                 f'https://rt.data.gov.hk/v2/transport/citybus/stop/{stop_id}') as request:
#             data = (await request.json())['data']
#             return {
#                 'zh': data.get('name_tc', '未有資料'),
#                 'en': data.get('name_en', 'N/A')
#             }

#     async with semaphore:
#         ends = await asyncio.gather(_route_ends(route, 'outbound', session),
#                                     _route_ends(route, 'inbound', session))

#     for direction in ends:
#         if direction is None:
#             continue
#         cache.setdefault(direction[0], await _stop_name(direction[0], session))
#         cache.setdefault(direction[1], await _stop_name(direction[1], session))

#     return route, {
#         'outbound': [] if ends[0] is None else [{
#             'id': f'{route}_outbound_1',
#             'description': None,
#             'orig': cache.get(ends[0][0]),
#             'dest': cache.get(ends[0][1]),
#         }],
#         'inbound': [] if ends[1] is None else [{
#             'id': f'{route}_inbound_1',
#             'description': None,
#             'orig': cache.get(ends[1][0]),
#             'dest': cache.get(ends[1][1]),

#         }]
#     }



import asyncio
import csv
from datetime import datetime, timedelta
from typing import Generator

import aiohttp
import pytz

from . import t
from ._utils import dt_to_8601, ensure_session, error_eta, search_location


@ensure_session
async def routes(*, session: aiohttp.ClientSession) -> dict[str, t.Route]:
    routes_ = {}
    async with session.get(
            'https://opendata.mtr.com.hk/data/light_rail_routes_and_stops.csv') as request:
        for row in csv.reader((await request.text('utf-8')).splitlines()[1:]):
            # column definition:
            #   route, direction, stopCode, stopID, stopTCName, stopENName, seq
            direction = 'outbound' if row[1] == '1' else 'inbound'
            routes_.setdefault(row[0], {'outbound': [], 'inbound': []})

            if row[6] == '1.00':
                # original
                routes_[row[0]][direction].append({
                    'id': None,
                    'description': None,
                    'orig':  {'en': row[5], 'tc': row[4]},
                    'dest': {}
                })
            else:
                # destination
                if row[0] in ('705', '706'):
                    routes_[row[0]][direction][0]['id'] =\
                        f'{row[0]}_{direction}_TSW Circular'
                else:
                    routes_[row[0]][direction][0]['id'] =\
                        f'{row[0]}_{direction}_{row[5]}'

                routes_[row[0]][direction][0]['dest'] = {
                    'en': row[5], 'tc': row[4]
                }
    return routes_


@ ensure_session
async def stops(route_id: str, *, session: aiohttp.ClientSession) -> Generator[t.Stop, None, None]:
    async with session.get(
            'https://opendata.mtr.com.hk/data/light_rail_routes_and_stops.csv') as request:
        # pylint: disable=line-too-long
        stops_ = [stop for stop in csv.reader((await request.text('utf-8')).splitlines()[1:])
                  if set((stop[0], 'outbound' if stop[1] == '1' else 'inbound')) == set(route_id.split('_')[:2])]

    if len(stops_) == 0:
        raise KeyError('route not exists')

    locations = await asyncio.gather(
        *[search_location(f'\u8f15\u9435\uff0d{s[4]}', session) for s in stops_])
    return ({
        'id': s[3],
        'seq': int(s[6].removesuffix('.00')),
        'name': {'tc': s[4], 'en': s[5]},
        'location': locations[i]
    } for i, s in enumerate(stops_))


@ ensure_session
async def etas(route_id: str,
               stop_id: str,
               language: t.Language = 'tc',
               *,
               session: aiohttp.ClientSession) -> t.Etas:
    route, _, destination = route_id.split('_')
    lc = 'ch' if language == 'tc' else 'en'

    async with session.get('https://rt.data.gov.hk/v1/transport/mtr/lrt/getSchedule',
                           params={'station_id': stop_id}) as request:
        response = await request.json()

    if len(response) == 0 or response.get('status', 0) == 0:
        return error_eta('api-error')
    if all(platform.get('end_service_status', False)
            for platform in response['platform_list']):
        return error_eta('eos')

    etas_ = []
    cnt_stopped = 0
    timestamp = datetime.fromisoformat(response['system_time'])\
        .astimezone(pytz.timezone('Asia/Hong_kong'))

    for platform in response['platform_list']:
        for eta in platform.get('route_list', []):
            # NOTE: 751P have no destination and eta
            if eta['route_no'] != route:
                continue
            if eta.get('stop') == 1:
                cnt_stopped += 1
                continue
            if eta['dest_en'] != destination:
                continue

            eta_min: str = eta[f'time_{lc}'].split(' ')[0]  # e.g. 3 分鐘 / 即將抵達
            if eta_min.isnumeric():
                etas_.append({
                    'eta': dt_to_8601(
                        timestamp + timedelta(minutes=float(eta_min))),
                    'is_arriving': False,
                    'is_scheduled': False,
                    'extras': {
                        'destination': eta[f'dest_{lc}'],
                        'varient': None,
                        'platform': str(platform['platform_id']),
                        'car_length': eta['train_length']
                    },
                })
            else:
                etas_.append({
                    'eta': dt_to_8601(timestamp),
                    'is_arriving': True,
                    'is_scheduled': False,
                    'remark': eta_min,
                    'extras': {
                        'destination': eta[f'dest_{lc}'],
                        'varient': None,
                        'platform': str(platform['platform_id']),
                        'car_length': eta['train_length']
                    }
                })

    if len(etas_) > 0:
        return {
            'timestamp': dt_to_8601(timestamp),
            'message': None,
            'etas': etas_
        }
    if 'red_alert_status' in response.keys():
        return error_eta(response[f'red_alert_message_{lc}'])
    if cnt_stopped > 0:
        return error_eta('eos')
    return error_eta('empty')



from datetime import datetime
from typing import Generator

import aiohttp
import bs4
import pytz

from . import t
from ._utils import dt_to_8601, ensure_session, error_eta, ua_header


@ensure_session
async def routes(*, session: aiohttp.ClientSession) -> dict[str, t.Route]:
    def description(route: dict[str,]):
        descr = {}
        for lc, routes_ in descriptions.items():
            for service in routes_[route['routeNo']]:
                if service['route_name'] == route[f'routeName_{lc}']:
                    descr[lc] = service['description']
                    break
        return descr or None

    descriptions = {'e': {}, 'c': {}}
    async with aiohttp.ClientSession(headers=ua_header()) as sess_cua:
        for lc in ('e', 'c'):
            async with sess_cua.get(
                    f'https://www.nlb.com.hk/language/set/{"en" if lc == "e" else "zh"}'):
                pass
            async with sess_cua.get('https://www.nlb.com.hk/route') as request:
                bs = bs4.BeautifulSoup(await request.text(), "html.parser")

            for tr in bs.select('table.property-table tr')[1:]:
                route_no = tr.select('td')[0].get_text(strip=True)

                descriptions[lc].setdefault(route_no, [])
                descriptions[lc][route_no].append({
                    'route_name': tr.select('td')[1].contents[1].get_text(),
                    'description': tr.select('td')[1].contents[2].get_text(strip=True)
                })

    routes_ = {}
    async with session.get(
            'https://rt.data.gov.hk/v2/transport/nlb/route.php?action=list') as request:
        for route in (await request.json())['routes']:
            routes_.setdefault(route['routeNo'],
                               {'outbound': [], 'inbound': []})
            direction = ('inbound'
                         if len(routes_[route['routeNo']]['outbound'])
                         else 'outbound')
            detail = {
                'description': description(route),
                'orig': {
                    'tc': route['routeName_c'].split(' \u003E ')[0],
                    'en': route['routeName_e'].split(' \u003E ')[0],
                },
                'dest': {
                    'tc': route['routeName_c'].split(' \u003E ')[1],
                    'en': route['routeName_e'].split(' \u003E ')[1],
                }
            }

            # when both the `outbound` and `inbound` have data, it is a special route.
            if all(len(b) for b in routes_[route['routeNo']].values()):
                for bound, parent_rt in routes_[route['routeNo']].items():
                    for r in parent_rt:
                        # special routes usually only differ from either orig or dest stop
                        if (r['orig']['en'] == detail['orig']['en']
                                or r['dest']['en'] == detail['dest']['en']):
                            direction = bound
                            break
                    else:
                        continue
                    break

            routes_[route['routeNo']][direction].append({
                'id': f'{route["routeNo"]}_{direction}_{route["routeId"]}',
                **detail
            })
    return routes_


@ensure_session
async def stops(route_id: str, *, session: aiohttp.ClientSession) -> Generator[t.Stop, None, None]:
    # pylint: disable=line-too-long
    async with session.get(
            f'https://rt.data.gov.hk/v2/transport/nlb/stop.php?action=list&routeId={route_id.split("_")[-1]}') as request:
        if len(stops_ := (await request.json())['stops']) == 0:
            raise KeyError('route not exists')

    return ({
        'id': stop['stopId'],
        'seq': idx,
        'name': {
            'tc': stop['stopName_c'],
            'en': stop['stopName_e']
        },
        'location': (stop['latitude'], stop['longitude'])
    } for idx, stop in enumerate(stops_))


@ensure_session
async def etas(route_id: str,
               stop_id: str,
               language: t.Language = 'tc',
               *,
               session: aiohttp.ClientSession) -> t.Etas:
    async with session.get('https://rt.data.gov.hk/v2/transport/nlb/stop.php',
                           params={
                               'action': 'estimatedArrivals',
                               'routeId': route_id.split('_')[-1],
                               'stopId': stop_id,
                               'language': language,
                           }) as request:
        response = await request.json()

    if len(response) == 0:
        # incorrect parameter will result in a empty json response
        return error_eta('api-error')
    if not response.get('estimatedArrivals', []):
        return error_eta('empty')

    etas_ = []
    timestamp = datetime.now().replace(tzinfo=pytz.timezone('Etc/GMT-8'))

    for eta in response['estimatedArrivals']:
        eta_dt = datetime.fromisoformat(eta['estimatedArrivalTime']) \
            .astimezone(pytz.timezone('Asia/Hong_kong'))

        etas_.append({
            'eta': dt_to_8601(eta_dt),
            'is_arriving': (eta_dt - timestamp).total_seconds() < 60,
            'is_scheduled': not (eta.get('departed') == '1'
                                 and eta.get('noGPS') == '1'),
            'extras': {
                'destinaion': None,
                'varient': eta.get('routeVariantName'),
                'platform': None,
                'car_length': None
            },
            'remark': None,
        })

    return {
        'timestamp': dt_to_8601(timestamp),
        'message': None,
        'etas': etas_
    }



import csv
import json
import tempfile
import xml.etree.ElementTree as ET
from pathlib import Path
from typing import Literal, Union

import aiohttp

from ._utils import ensure_session, is_up_to_date

_BASE_PATH = Path(tempfile.gettempdir())


def _bound_id_conv(bound_id: Literal['1', '2']):
    return 'outbound' if bound_id == '1' else 'inbound'


@ensure_session
async def journey_time(*, session: aiohttp.ClientSession):
    path = _BASE_PATH.joinpath('_hketa_rb.json')
    # path = Path(tempfile.gettempdir()).joinpath('_hketa_rb.xml')

    if (path.exists()
            and await is_up_to_date(path,
                                    'https://static.data.gov.hk/td/routes-fares-xml/DATA_LAST_UPDATED_DATE.csv',
                                    session)):
        with open(path, 'r', encoding='utf-8') as f:
            return json.load(f)

    async with session.get('https://static.data.gov.hk/td/routes-fares-xml/ROUTE_BUS.xml') as request:
        data = {}
        for r in ET.fromstring(await request.text()).iter('ROUTE'):
            co = r.find('COMPANY_CODE').text.lower().split('+')
            r_name = r.find('ROUTE_NAMEC').text

            for c in co:
                data.setdefault(c, {})
                data[c].setdefault(r_name, [])
                data[c][r_name].append({
                    'td_route_id': r.find('ROUTE_ID').text,
                    'orig': r.find('LOC_START_NAMEC').text,
                    'dest': r.find('LOC_END_NAMEC').text,
                    'time': r.find('JOURNEY_TIME').text,
                })

    path.write_text(json.dumps(data, ensure_ascii=False), encoding='utf-8')
    return data


@ensure_session
async def gtfs_routes(*, session: aiohttp.ClientSession):
    path = _BASE_PATH.joinpath('_hketa_gtfs_routes.json')

    if path.exists() and await is_up_to_date(path, 'https://static.data.gov.hk/td/pt-headway-en/DATA_LAST_UPDATED_DATE.csv', session):
        with open(path, 'r', encoding='utf-8') as f:
            return json.load(f)

    routes = {}
    async with session.get('https://static.data.gov.hk/td/pt-headway-tc/routes.txt') as request:
        for line in (l for l in csv.reader((await request.text()).splitlines()[1:]) if l[4] == '3'):
            for co in line[1].lower().split('+'):
                routes.setdefault(co, {})
                routes[co].setdefault(line[2], [])
                routes[co][line[2]].append({
                    'id': line[0],
                    **dict(zip(('orig', 'dest'), line[3].replace('(循環線)', '').split(' - ')))
                })

    with open(path, 'w', encoding='utf-8') as f:
        json.dump(routes, f)


@ensure_session
async def gtfs_calendar(*, session: aiohttp.ClientSession) -> dict[str, Union[str, list[str]]]:
    path = _BASE_PATH.joinpath('_hketa_gtfs_calendar.json')

    if path.exists() and await is_up_to_date(path, 'https://static.data.gov.hk/td/pt-headway-en/DATA_LAST_UPDATED_DATE.csv', session):
        with open(path, 'r', encoding='utf-8') as f:
            return json.load(f)

    async with session.get('https://static.data.gov.hk/td/pt-headway-tc/calendar.txt') as request:
        calendar = {
            c[0]: {
                'weekday': tuple(1 if d == "1" else 0 for d in c[1:8]),
                'incl': [],
                'excl': []
            } for c in csv.reader((await request.text()).splitlines()[1:])
        }

    async with session.get('https://static.data.gov.hk/td/pt-headway-tc/calendar_dates.txt') as request:
        for d in csv.reader((await request.text()).splitlines()[1:]):
            calendar[d[0]]['incl' if d[2] == '1' else 'excl'].append(d[1])

    with open(path, 'w', encoding='utf-8') as f:
        json.dump(calendar, f)

    return calendar


@ensure_session
async def gtfs_frequencies(*, session: aiohttp.ClientSession):
    path = _BASE_PATH.joinpath('_hketa_gtfs_freq.json')
    # schedules = {}
    # async with session.get('https://static.data.gov.hk/td/pt-headway-tc/trips.txt') as request:
    #     trips = {}
    #     for t in (l[-1].split('-') for l in csv.reader((await request.text()).splitlines()[1:])):
    #         direction = 'outbound' if t[1] == '1' else 'inbound'
    #         trips.setdefault(t[0], {})
    #         trips[t[0]].setdefault(direction, {})
    #         trips[t[0]][direction].setdefault(t[2], [])
    #         trips[t[0]][direction][t[2]].append(t[3])

    async with session.get('https://static.data.gov.hk/td/pt-headway-tc/frequencies.txt') as request:
        freqs = {}
        for freq in csv.reader((await request.text()).splitlines()[1:]):
            rid, bound, sid, _ = freq[0].split('-')
            bound = _bound_id_conv(bound)

            freqs.setdefault(rid, {'outbound': {}, 'inbound': {}})
            freqs[rid][bound].setdefault(sid, [])
            freqs[rid][bound][sid].append({
                'start': freq[1],
                'end': freq[2],
                'interval': freq[3]
            })

    with open(path, 'w', encoding='utf-8') as f:
        json.dump(freqs, f)


@ensure_session
async def gtfs_fares(*, session: aiohttp.ClientSession):
    path = _BASE_PATH.joinpath('_hketa_gtfs_fares.json')

    if path.exists() and await is_up_to_date(path, 'https://static.data.gov.hk/td/pt-headway-en/DATA_LAST_UPDATED_DATE.csv', session):
        with open(path, 'r', encoding='utf-8') as f:
            return json.load(f)

    async with session.get('https://static.data.gov.hk/td/pt-headway-tc/fare_attributes.txt') as request:
        fares = {}
        for fare in csv.reader((await request.text()).splitlines()[1:]):
            rid, bound, idx_on, idx_off = fare[0].split('-')

            if int(idx_off) - int(idx_on) != 1:
                continue

            fares.setdefault(rid, {'outbound': [], 'inbound': []})
            fares[rid][_bound_id_conv(bound)].append(fare[1])

    with open(path, 'w', encoding='utf-8') as f:
        json.dump(fares, f)


@ensure_session
async def gtfs_stops(*, session: aiohttp.ClientSession):
    path = _BASE_PATH.joinpath('_hketa_gtfs_stops.json')

    if path.exists() and await is_up_to_date(path, 'https://static.data.gov.hk/td/pt-headway-en/DATA_LAST_UPDATED_DATE.csv', session):
        with open(path, 'r', encoding='utf-8') as f:
            return json.load(f)

    def parse_name(names: str):
        parsed = {}
        for name in [n for n in names.split('|') if '+' not in n]:
            if name[0] != '[':
                parsed['_'] = name
            else:
                co = name[name.find('[') + 1:name.find(']')]\
                    .lower()\
                    .replace('lwb', 'kmb')
                parsed[co] = name[name.find(']') + 1:]
        # {co: gtfs_name for name in names.split('|') for co, gtfs_name in [name.split(' ', 1)]}
        return parsed

    async with session.get('https://static.data.gov.hk/td/pt-headway-tc/stops.txt') as request:
        stops = {
            s[0]: {
                'name': parse_name(s[1]),
                'lat': s[2],
                'lng': s[3]
            } for s in csv.reader((await request.text()).splitlines()[1:])
        }

    with open(path, 'w', encoding='utf-8') as f:
        json.dump(stops, f)



from pathlib import Path
from setuptools import find_packages, setup

with open(Path(__file__).parent.joinpath('README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='hketa',
    version='v0.0.7',
    description='A Python package that provides normalised data access to public transports in Hong Kong.',
    long_description_content_type='text/markdown',
    long_description=long_description,
    url="https://github.com/SuperDumbTM/hketa",
    install_requires=[
        'aiohttp',
        'beautifulsoup4',
        'pyproj',
        'pytz'
    ],
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    include_package_data=True
)



import asyncio
from datetime import datetime
from itertools import chain
from typing import Generator, Literal, Optional

import aiohttp

from . import t
from ._utils import dt_to_8601, ensure_session, error_eta


@ensure_session
async def routes(*, session: aiohttp.ClientSession) -> dict[str, t.Route]:
    routes_ = {}
    specials = set()

    async with session.get('https://data.etabus.gov.hk/v1/transport/kmb/route') as request:
        for route in (await request.json())['data']:
            routes_.setdefault(route['route'], {'inbound': [], 'outbound': []})
            direction = 'outbound' if route['bound'] == 'O' else 'inbound'

            routes_[route['route']][direction].append({
                'id': f'{route["route"]}_{direction}_{route["service_type"]}',
                'description': None,
                'orig': {'tc': route['orig_tc'], 'en': route['orig_en']},
                'dest': {'tc': route['dest_tc'], 'en': route['dest_en']},
            })

            if len(routes_[route['route']][direction]) > 1:
                specials.add(
                    (route['route'], '1' if route['bound'] == 'O' else '2'))

    varients = chain(*(await asyncio.gather(*[_variants(r, d, session) for r, d in specials])))
    for varient in (v for v in varients if v['ServiceType'] != '01   '):
        # pylint: disable=line-too-long
        for service in routes_[varient['Route']]['outbound' if varient['Bound'] == '1' else 'inbound']:
            if service['id'].split('_')[2] == varient['ServiceType'].strip().removeprefix('0'):
                service['description'] = {
                    'tc': varient['Desc_CHI'],
                    'en': varient['Desc_ENG']
                }
                break
    return routes_


@ensure_session
async def stops(route_id: str, *, session: aiohttp.ClientSession) -> list[dict[str,]]:
    async def fetch(stop: dict, session: aiohttp.ClientSession):
        async with session.get(
                f'https://data.etabus.gov.hk/v1/transport/kmb/stop/{stop["stop"]}') as request:
            detail = (await request.json())['data']
            return {
                'id': stop['stop'],
                'seq': int(stop['seq']),
                'name': {
                    'tc': detail.get('name_tc'),
                    'en': detail.get('name_en'),
                },
                'location': (detail['lat'], detail['long'])
            }

    # pylint: disable=line-too-long
    async with session.get(
            f'https://data.etabus.gov.hk/v1/transport/kmb/route-stop/{"/".join(route_id.split("_"))}') as request:
        data = await asyncio.gather(
            *[fetch(stop, session) for stop in (await request.json())['data']])

    if len(data) == 0:
        raise KeyError('route not exists')
    return data


@ensure_session
async def etas(route_id: str,
               stop_id: str,
               language: t.Language = 'tc',
               *,
               session: aiohttp.ClientSession) -> t.Etas:
    route, direction, service_type = route_id.split('_')

    # pylint: disable=line-too-long
    async with session.get(
            f'https://data.etabus.gov.hk/v1/transport/kmb/eta/{stop_id}/{route}/{service_type}') as request:
        response = await request.json()

    if len(response) == 0:
        return error_eta('api-error', language=language)
    if response.get('data') is None:
        return error_eta('empty', language=language)

    etas_ = []
    timestamp = datetime.fromisoformat(response['generated_timestamp'])

    for eta in response['data']:
        if eta['dir'].lower() != direction[0]:
            continue
        if eta['eta'] is None:
            if eta['rmk_en'] == 'The final bus has departed from this stop':
                return error_eta('eos')
            elif eta['rmk_en'] == '':
                return error_eta('empty')
            return error_eta(eta[f'rmk_{language}'])

        eta_dt = datetime.fromisoformat(eta['eta'])
        etas_.append({
            'eta': dt_to_8601(eta_dt),
            'is_arriving': (eta_dt - timestamp).total_seconds() < 30,
            'is_scheduled': eta.get(f'rmk_{language}') in ('\u539f\u5b9a\u73ed\u6b21', 'Scheduled Bus'),
            'extras': {
                'destinaion': eta[f'dest_{language}'],
                'varient': _varient_text(eta['service_type'], language),
                'platform': None,
                'car_length': None
            },
            'remark': eta[f'rmk_{language}'],
        })

    return {
        'timestamp': dt_to_8601(timestamp),
        'message': None,
        'etas': etas_
    }


async def _variants(route: str,
                    direction: Literal['1', '2'],
                    session: aiohttp.ClientSession) -> list[dict]:
    async with session.request('GET',
                               'https://search.kmb.hk/KMBWebSite/Function/FunctionRequest.ashx',
                               params={
                                   'action': 'getSpecialRoute',
                                   'route': route,
                                   'bound': direction
                               },
                               ) as requset:
        return (await requset.json(content_type=None))['data']['routes']


def _varient_text(service_type: str, language: t.Language) -> Optional[str]:
    if service_type == '1':
        return None
    return '\u7279\u5225\u73ed\u6b21' if language == 'tc' else 'Special Departure'



import csv
from datetime import datetime, timedelta
from typing import Generator

import aiohttp
import pytz

from . import t
from ._utils import dt_to_8601, ensure_session, error_eta


@ensure_session
async def routes(*, session: aiohttp.ClientSession) -> dict[str, t.Route]:
    routes_ = {}
    async with session.get('https://opendata.mtr.com.hk/data/mtr_bus_stops.csv') as request:
        for row in csv.reader((await request.text('utf-8')).splitlines()[1:]):
            # column definition:
            #   route, direction, seq, stopID, stopLAT, stopLONG, stopTCName, stopENName
            direction = 'outbound' if row[1] == 'O' else 'inbound'
            routes_.setdefault(row[0], {'outbound': [], 'inbound': []})

            if row[2] == '1.00' or row[2] == '1':
                # orignal
                routes_[row[0]][direction].append({
                    'id': f'{row[0]}_{direction}_1',
                    'description': None,
                    'orig': {'en': row[7], 'tc': row[6]},
                    'dest': {}
                })
            else:
                # destination
                routes_[row[0]][direction][0]['dest'] = {
                    'en': row[7], 'tc': row[6]
                }
    return routes_


@ensure_session
async def stops(route_id: str, *, session: aiohttp.ClientSession) -> Generator[t.Stop, None, None]:
    async with session.get('https://opendata.mtr.com.hk/data/mtr_bus_stops.csv') as request:
        # pylint: disable=line-too-long
        stops_ = [stop for stop in csv.reader((await request.text('utf-8')).splitlines()[1:])
                  if set((stop[0], 'outbound' if stop[1] == 'O' else 'inbound')) == set(route_id.split('_')[:2])]

    if len(stops_) == 0:
        raise KeyError('route not exists')
    return ({
        'id': s[3],
        'seq': int(s[2].removesuffix('.00')),
        'name': {'tc': s[6], 'en': s[7]},
        'location': (s[4], s[5])
    } for s in stops_)


@ ensure_session
async def etas(route_id: str,
               stop_id: str,
               language: t.Language = 'tc',
               *,
               session: aiohttp.ClientSession) -> t.Etas:
    route, *_ = route_id.split('_')

    async with session.post('https://rt.data.gov.hk/v1/transport/mtr/bus/getSchedule',
                            json={'routeName': route, 'language': language}) as request:
        response = await request.json()

    if len(response) == 0:
        return error_eta('api-error')
    if response['routeStatusRemarkTitle'] is not None:
        if response['routeStatusRemarkTitle'] in ('\u505c\u6b62\u670d\u52d9', 'Non-service hours'):
            return error_eta('eos')
        return error_eta(response['routeStatusRemarkTitle'])

    etas_ = []
    timestamp = datetime.strptime(response['routeStatusTime'], '%Y/%m/%d %H:%M')\
        .astimezone(pytz.timezone('Asia/Hong_kong'))

    for stop in (s for s in response['busStop'] if s['busStopId'] == stop_id):
        for eta in stop['bus']:
            time_ref = 'departure' if eta['arrivalTimeInSecond'] == '108000' else 'arrival'
            if any(c.isdigit() for c in eta[f'{time_ref}TimeText']):
                # eta TimeText has numbers (e.g. 3 分鐘/3 Minutes)
                etas_.append({
                    'eta': dt_to_8601(
                        timestamp + timedelta(seconds=int(eta[f'{time_ref}TimeInSecond']))),
                    'is_arriving': False,
                    'is_scheduled': eta['busLocation']['longitude'] == 0,
                    'extras': {
                        'destination': None,
                        'varient': None,
                        'platform': None,
                        'car_length': None
                    },
                    'remark': None
                })
            else:
                etas_.append({
                    'eta': dt_to_8601(timestamp),
                    'is_arriving': True,
                    'is_scheduled': eta['busLocation']['longitude'] == 0,
                    'extras': {
                        'destination': None,
                        'varient': None,
                        'platform': None,
                        'car_length': None
                    },
                    'remark': None,
                })

    return {
        'timestamp': dt_to_8601(timestamp),
        'message': None,
        'etas': etas_
    }
